import json
from enum import Enum
from json import JSONEncoder


class Encoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Arity):
            return o.lookup()
        return json.JSONEncoder.default(self, o)


class Arity(Enum):
    SOLO = 1
    DUO = 2
    TRIO = 3

    def lookup(self):
        dict = {
            Arity.SOLO: 'Solo',
            Arity.DUO: 'Duo',
            Arity.TRIO: 'Trio',
        }
        return dict[self]

    def __str__(self):
        return self.toJSON()

    def toJSON(self):
        return json.dumps(self, indent=2, cls=Encoder)

    def __lt__(self, other):
        return self.value < other.value



from flask import Blueprint, render_template, request
from werkzeug.exceptions import abort

from osuca.db import get_db

bp = Blueprint('combination', __name__)


@bp.route("/combination/<key>/<reverse>", methods=['GET', 'POST'])
def combination(key, reverse):
    db = get_db()

    query_result = db.course_combination_aggregate()
    selection_label = "All Courses"
    if request.method == 'POST' and request.form['course'] != "All Courses":
        selection_label = request.form['course']
        selection = db.course(selection_label)
        query_result = db.course_combination_aggregate(selection)

    def course(cca):
        return cca[0][0]

    def count(cca):
        return cca[1].count

    def mean(cca):
        return cca[1].mean
    dispatcher = {'course': course, 'count': count, 'mean': mean}

    return render_template("combination.html",
                           selection_label=selection_label,
                           course=sorted(db.course()),
                           query_result=sorted(query_result,
                                               key=dispatcher[key],
                                               reverse=(reverse == 'True')))



import csv
from functools import partial

from osuca.model.course import Course
from osuca.model.quarter import Quarter
from osuca.model.arity import Arity


class Aggregate():
    def __init__(self):
        self.count = 0
        self.sum = 0
        self.mean = None

    def compute(self):
        if self.count != 0:
            self.mean = self.sum / self.count


class Analytics():
    """Contains in-memory structures for analytics."""

    def __init__(self, input):
        print('Initializing analytics.')
        self.evaluation = set()

        reader = csv.reader(input.splitlines(), delimiter=',')
        next(reader)  # skip header

        id = 0
        for row in reader:
            quarter = Quarter(row[5])
            if quarter.invalid():
                print('Rejecting line %d...' % id)
            else:
                id += 1
                term = quarter.term
                year = quarter.year
                if row[11].lower() == 'yes':
                    ca = Course(row[1])
                    cb = Course(row[7])
                    cc = Course(row[12])
                    da = int(row[2])
                    db = int(row[8])
                    dc = int(row[13])
                    trio = frozenset({ca, cb, cc})
                    self.evaluation.add(
                        (id, ca, term, year, da, trio, Arity(Arity.TRIO)))
                    self.evaluation.add(
                        (id, cb, term, year, db, trio, Arity(Arity.TRIO)))
                    self.evaluation.add(
                        (id, cc, term, year, dc, trio, Arity(Arity.TRIO)))
                elif row[6].lower() == 'yes':
                    ca = Course(row[1])
                    cb = Course(row[7])
                    da = int(row[2])
                    db = int(row[8])
                    duo = frozenset({ca, cb})
                    self.evaluation.add(
                        (id, ca, term, year, da, duo, Arity(Arity.DUO)))
                    self.evaluation.add(
                        (id, cb, term, year, db, duo, Arity(Arity.DUO)))
                else:
                    ca = Course(row[1])
                    da = int(row[2])
                    solo = frozenset({ca})
                    self.evaluation.add(
                        (id, ca, term, year, da, solo, Arity(Arity.SOLO)))

        print('response count: %d' % len(self.response()))
        print('evaluation count: %d' % self.evaluation_count())
        print('course count: %d' % len(self.course()))
        print('term count: %d' % len(self.term()))
        print('year count: %d' % len(self.year()))
        print('quarter count: %d' % len(self.quarter()))
        print('course-term count: %d' % len(self.course_term()))
        print('course-year count: %d' % len(self.course_year()))
        print('course-quarter count: %d' % len(self.course_quarter()))
        print('combination count: %d' % len(self.combination()))
        print('solo count: %d' % len(self.solo()))
        print('duo count: %d' % len(self.duo()))
        print('trio count: %d' % len(self.trio()))

    def aggregate(self, o, key=None):
        agg = Aggregate()
        if key is None:
            for e in self.evaluation:
                agg.sum = agg.sum + e[4]
                agg.count += 1
        elif len(key) == 1:
            for e in self.evaluation:
                if e[key[0]] == o:
                    agg.sum = agg.sum + e[4]
                    agg.count += 1
        elif len(key) == 2:
            if key[0] == 2 and key[1] == 3:            # quarter, i.e, term-year
                for e in self.evaluation:
                    if e[2] == o.term and e[3] == o.year:
                        agg.sum = agg.sum + e[4]
                        agg.count += 1
            elif key[0] == 1 and key[1] == 2:          # course-term
                for e in self.evaluation:
                    if e[1] == o[0] and e[2] == o[1]:
                        agg.sum = agg.sum + e[4]
                        agg.count += 1
            elif key[0] == 1 and key[1] == 3:          # course-year
                for e in self.evaluation:
                    if e[1] == o[0] and e[3] == o[1]:
                        agg.sum = agg.sum + e[4]
                        agg.count += 1
            elif key[0] == 1 and key[1] == 5:          # course-combination
                for e in self.evaluation:
                    if e[1] == o[0] and e[5] == o[1]:
                        agg.sum = agg.sum + e[4]
                        agg.count += 1
        elif key[0] == 1 and key[1] == 2 and key[2] == 3:  # course-term-year
            for e in self.evaluation:
                if e[1] == o[0] and e[2] == o[1].term and e[3] == o[1].year:
                    agg.sum = agg.sum + e[4]
                    agg.count += 1

        agg.compute()
        return (o,) + (agg,)

    def evaluation_count(self):
        return len(self.evaluation)

    def response(self):
        """Unary predicate for response."""
        return {e[0] for e in self.evaluation}

    def response_aggregate(self):
        """Unary predicate for representing aggregate response."""
        return set(map(self.aggregate, self.response()))

    def course(self, label=None):
        """Unary predicate for course."""
        if label is None:
            return {e[1] for e in self.evaluation}

        tokens = label.split('-')
        subject = tokens[0].split(' ')[0].strip()
        id = tokens[0].split(' ')[1].strip()
        return {e[1] for e in self.evaluation if e[1].subject == subject and e[1].id == id}

    def course_aggregate(self):
        """Binary predicate for course-aggregate relation."""
        return set(map(partial(self.aggregate, key=(1,)), self.course()))

    def term(self):
        """Unary predicate for term."""
        return {e[2] for e in self.evaluation}

    def term_aggregate(self):
        """Binary predicate for term-aggregate relation."""
        return set(map(partial(self.aggregate, key=(2,)), self.term()))

    def arity(self):
        return {Arity(len(e[5])) for e in self.evaluation}

    def arity_aggregate(self):
        """Binary predicate for arity-aggregate relation."""
        return set(map(partial(self.aggregate, key=(6,)), self.arity()))

    def year(self):
        """Unary predicate for year."""
        return {e[3] for e in self.evaluation}

    def year_aggregate(self):
        """Binary predicate for year-aggregate relation."""
        return set(map(partial(self.aggregate, key=(3,)), self.year()))

    def quarter(self):
        """Unary predicate for quarter."""
        return {Quarter(e[2].lookup() + e[3]) for e in self.evaluation}

    def quarter_aggregate(self):
        """Binary predicate for quarter-aggregate relation."""
        return set(map(partial(self.aggregate, key=(2, 3)), self.quarter()))

    def course_term(self, course=None):
        """Binary predicate for course-term relation.

        Course is taken in Term, e.g., CS 261 is taken in Summer.
        Term has Course, e.g.,  Summer has CS 261. 
        """
        if course is None:
            return {(e[1], e[2]) for e in self.evaluation}
        else:
            return {(e[1], e[2]) for e in self.evaluation if e[1] in course}

    def course_arity(self, course=None):
        """Binary predicate for course-arity relation.

        Course is taken in Term, e.g., CS 261 is taken in Summer.
        Term has Course, e.g.,  Summer has CS 261. 
        """
        if course is None:
            return {(e[1], e[6]) for e in self.evaluation}
        else:
            return {(e[1], e[6]) for e in self.evaluation if e[1] in course}

    def course_arity_aggregate(self, course=None):
        """Ternary predicate for course-arity-aggregate relation."""
        course_arity = self.course_arity(course)
        return set(map(partial(self.aggregate, key=(1, 6)), course_arity))

    def course_term_aggregate(self, course=None):
        """Ternary predicate for course-term-aggregate relation."""
        course_term = self.course_term(course)
        return set(map(partial(self.aggregate, key=(1, 2)), course_term))

    def course_year(self):
        """Binary predicate for course-year relation.

        Course is taken in Year, e.g., CS 261 is taken in Summer.
        Year has Course, e.g.,  2022 has CS 261. 
        """
        return {(e[1], e[3]) for e in self.evaluation}

    def course_year_aggregate(self):
        """Ternary predicate for course-year-aggregate relation."""
        return set(map(partial(self.aggregate, key=(1, 3)), self.course_year()))

    def course_quarter(self):
        """Binary predicate for course-quarter relation.
        """
        return {(e[1], Quarter(e[2].lookup() + e[3])) for e in self.evaluation}

    def course_quarter_aggregate(self):
        """Ternary predicate for course-quarter-aggregate relation."""
        return set(map(partial(self.aggregate, key=(1, 2, 3)), self.course_quarter()))

    def combination(self):
        return {e[5] for e in self.evaluation}

    def solo(self):
        """Combination with a single course.
        """
        return {e[5] for e in self.evaluation if (len(e[5]) == 1)}

    def duo(self):
        """Combination with two courses.
        """
        return {e[5] for e in self.evaluation if (len(e[5]) == 2)}

    def trio(self):
        """Combination with three courses.
        """
        return {e[5] for e in self.evaluation if (len(e[5]) == 3)}

    def course_combination(self, course=None):
        """Binary predicate for course-combination relation.
        """
        if course is None:
            return {(e[1], e[5]) for e in self.evaluation}
        return {(e[1], e[5]) for e in self.evaluation if e[1] in course}

    def course_combination_aggregate(self, course=None):
        """Ternary predicate for course-combination-aggregate relation."""
        if course is None:
            return set(map(partial(self.aggregate, key=(1, 5)), self.course_combination()))
        return set(map(partial(self.aggregate, key=(1, 5)), self.course_combination(course)))



import json
from json import JSONEncoder

from osuca.model.analytics import Aggregate
from osuca.model.course import Course
from osuca.model.quarter import Term, Quarter
from osuca.model.arity import Arity


class JSONAdapter():
    def __init__(self, analytics):
        self.analytics = analytics

    def response(self, sort=False):
        """Returns json representation for all responses."""
        result = self.analytics.response()
        return sorted(list(result)) if sort else list(result)

    def response_aggregate(self):
        """Returns json representation for aggregate response."""
        # aggregate is the same for all response
        # we get the first tuple and then set the result to the second element
        return list(self.analytics.response_aggregate()).pop()[1]

    def course(self, sort=False):
        """Returns json representation for course predicate."""
        result = self.analytics.course()
        return sorted(list(result)) if sort else list(result)

    def course_aggregate(self, sort=False):
        """Returns json representation for course aggregate predicate."""
        result = [CourseAggregate(ca)
                  for ca in self.analytics.course_aggregate()]
        return sorted(result) if sort else result

    def term(self, sort=False):
        """Returns json representation for term predicate."""
        result = self.analytics.term()
        return sorted(list(result)) if sort else list(result)

    def arity(self, sort=False):
        """Returns json representation for term predicate."""
        result = self.analytics.arity()
        return sorted(list(result)) if sort else list(result)

    def term_aggregate(self, sort=False):
        """Returns json representation for term aggregate predicate."""
        result = [TermAggregate(ta) for ta in self.analytics.term_aggregate()]
        return sorted(result) if sort else result

    def arity_aggregate(self, sort=False):
        """Returns json representation for term aggregate predicate."""
        result = [ArityAggregate(aa)
                  for aa in self.analytics.arity_aggregate()]
        return sorted(result) if sort else result

    def year(self, sort=False):
        """Returns json representation for term predicate."""
        result = self.analytics.year()
        return sorted(list(result)) if sort else list(result)

    def year_aggregate(self, sort=False):
        """Returns json representation for year aggregate predicate."""
        result = [YearAggregate(ya) for ya in self.analytics.year_aggregate()]
        return sorted(result) if sort else result

    def quarter(self, sort=False):
        """Returns json representation for quarter predicate."""
        result = self.analytics.quarter()
        return sorted(list(result)) if sort else list(result)

    def quarter_aggregate(self, sort=False):
        """Returns json representation for quarter aggregate predicate."""
        result = [QuarterAggregate(qa)
                  for qa in self.analytics.quarter_aggregate()]
        return sorted(result) if sort else result

    def course_term(self, sort=False):
        """Returns json representation for course-term relations."""
        result = [CourseTerm(ct) for ct in self.analytics.course_term()]
        return sorted(result) if sort else result

    def course_term_aggregate(self, course=None, sort=False):
        """Returns json representation for course-term aggregate predicate."""
        if course is None:
            result = self.analytics.course_term_aggregate()
        else:
            result = self.analytics.course_term_aggregate(course)
        result = [CourseTermAggregate(cta) for cta in result]
        return sorted(result) if sort else result

    def course_arity_aggregate(self, course=None, sort=False):
        """Returns json representation for course-arity aggregate predicate."""
        if course is None:
            result = self.analytics.course_arity_aggregate()
        else:
            result = self.analytics.course_term_aggregate(course)
        result = [CourseArityAggregate(caa) for caa in result]
        return sorted(result) if sort else result

    def course_year(self, sort=False):
        """Returns json representation for course-year relations."""
        result = [CourseYear(cy) for cy in self.analytics.course_year()]
        result = sorted(result) if sort else result
        return sorted(result) if sort else result

    def course_year_aggregate(self, sort=False):
        """Returns json representation for course-term aggregate predicate."""
        result = self.analytics.course_year_aggregate()
        result = [CourseYearAggregate(cya) for cya in result]
        return sorted(result) if sort else result

    def course_quarter(self, sort=False):
        """Returns json representation for course-quarter relations."""
        result = [CourseQuarter(cq) for cq in self.analytics.course_quarter()]
        return sorted(result) if sort else result

    def course_quarter_aggregate(self, sort=False):
        """Returns json representation for course-term aggregate predicate."""
        result = self.analytics.course_quarter_aggregate()
        result = [CourseQuarterAggregate(cqa) for cqa in result]
        return sorted(result) if sort else result

    def course_combination_aggregate(self, sort=False):
        """Returns json representation for course-combination aggregate predicate."""
        result = self.analytics.course_combination_aggregate()
        result = [CourseCombinationAggregate(cqa) for cqa in result]
        return sorted(result) if sort else result


class Encoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Arity):
            return o.lookup()
        if isinstance(o, Course):
            return {"course": o.__dict__}
        if isinstance(o, Term):
            return o.lookup()
        if isinstance(o, Quarter):
            return {"quarter": o.__dict__}
        if isinstance(o, Aggregate):
            return {"aggregate": o.__dict__}
        if isinstance(o, CourseAggregate):
            return {"course": o.course.__dict__, "aggregate": o.aggregate.__dict__}
        if isinstance(o, ArityAggregate):
            return {"arity": o.arity.lookup(), "aggregate": o.aggregate.__dict__}
        if isinstance(o, TermAggregate):
            return {"term": o.term.lookup(), "aggregate": o.aggregate.__dict__}
        if isinstance(o, YearAggregate):
            return {"year": o.year, "aggregate": o.aggregate.__dict__}
        if isinstance(o, QuarterAggregate):
            return {
                "quarter": {
                    "term": o.quarter.term.lookup(),
                    "year": o.quarter.year
                },
                "aggregate": o.aggregate.__dict__
            }
        if isinstance(o, CourseTerm):
            return {"course": o.course.__dict__, "term": o.term.lookup()}
        if isinstance(o, CourseTermAggregate):
            return {
                "course": o.course.__dict__,
                "term": o.term.lookup(),
                "aggregate": o.aggregate.__dict__
            }
        if isinstance(o, CourseArityAggregate):
            return {
                "course": o.course.__dict__,
                "arity": o.arity.lookup(),
                "aggregate": o.aggregate.__dict__
            }
        if isinstance(o, CourseYear):
            return {"course": o.course.__dict__, "year": o.year}
        if isinstance(o, CourseYearAggregate):
            return {
                "course": o.course.__dict__,
                "year": o.year,
                "aggregate": o.aggregate.__dict__
            }
        if isinstance(o, CourseQuarter):
            return {
                "course": o.course.__dict__,
                "quarter": {
                    "term": o.quarter.term.lookup(),
                    "year": o.quarter.year
                }
            }
        if isinstance(o, CourseQuarterAggregate):
            return {
                "course": o.course,
                "quarter": o.quarter,
                "aggregate": o.aggregate
            }
        if isinstance(o, CourseCombinationAggregate):
            combo = {}
            for count, course in enumerate(o.combination):
                if course != o.course:
                    combo["course-" + str(count)] = (course.__dict__)
            return {
                "course": o.course.__dict__,
                "combination": combo,
                "aggregate": o.aggregate.__dict__
            }
        print(type(o))
        return json.JSONEncoder.default(self, o)


class CourseAggregate():
    def __init__(self, t):
        self.course = t[0]
        self.aggregate = t[1]

    def __lt__(self, other):
        return self.aggregate.mean < other.aggregate.mean


class ArityAggregate():
    def __init__(self, t):
        self.arity = t[0]
        self.aggregate = t[1]

    def __lt__(self, other):
        return self.arity < other.arity


class TermAggregate():
    def __init__(self, t):
        self.term = t[0]
        self.aggregate = t[1]

    def __lt__(self, other):
        return self.term < other.term


class YearAggregate():
    def __init__(self, t):
        self.year = t[0]
        self.aggregate = t[1]

    def __lt__(self, other):
        return self.year < other.year


class QuarterAggregate():
    def __init__(self, t):
        self.quarter = t[0]
        self.aggregate = t[1]

    def __lt__(self, other):
        return self.quarter < other.quarter


class CourseTerm():
    def __init__(self, t):
        self.course = t[0]
        self.term = t[1]

    def __lt__(self, other):
        if self.course < other.course:
            return True
        elif self.course > other.course:
            return False
        else:
            return self.term < other.term


class CourseTermAggregate():
    def __init__(self, t):
        self.course = t[0][0]
        self.term = t[0][1]
        self.aggregate = t[1]

    def __lt__(self, other):
        if self.course < other.course:
            return True
        elif self.course > other.course:
            return False
        else:
            return self.term < other.term


class CourseArity():
    def __init__(self, t):
        self.course = t[0]
        self.arity = t[1]

    def __lt__(self, other):
        if self.course < other.course:
            return True
        elif self.course > other.course:
            return False
        else:
            return self.arity < other.arity


class CourseArityAggregate():
    def __init__(self, t):
        self.course = t[0][0]
        self.arity = t[0][1]
        self.aggregate = t[1]

    def __lt__(self, other):
        if self.course < other.course:
            return True
        elif self.course > other.course:
            return False
        else:
            return self.arity < other.arity


class CourseYear():
    def __init__(self, t):
        self.course = t[0]
        self.year = t[1]

    def __lt__(self, other):
        if self.course < other.course:
            return True
        elif self.course > other.course:
            return False
        else:
            return self.year < other.year


class CourseYearAggregate():
    def __init__(self, t):
        self.course = t[0][0]
        self.year = t[0][1]
        self.aggregate = t[1]

    def __lt__(self, other):
        if self.course < other.course:
            return True
        elif self.course > other.course:
            return False
        else:
            return self.year < other.year


class CourseQuarter():
    def __init__(self, t):
        self.course = t[0]
        self.quarter = t[1]

    def __lt__(self, other):
        if self.course < other.course:
            return True
        elif self.course > other.course:
            return False
        else:
            return self.quarter < other.quarter


class CourseQuarterAggregate():
    def __init__(self, t):
        self.course = t[0][0]
        self.quarter = t[0][1]
        self.aggregate = t[1]

    def __lt__(self, other):
        if self.course < other.course:
            return True
        elif self.course > other.course:
            return False
        else:
            return self.quarter < other.quarter


class CourseCombinationAggregate():
    def __init__(self, t):
        self.course = t[0][0]
        self.combination = t[0][1]
        self.aggregate = t[1]

    def __lt__(self, other):
        if self.course < other.course:
            return True
        elif self.course > other.course:
            return False
        else:
            return self.aggregate.mean < other.aggregate.mean



import json
from json import JSONEncoder


class Encoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Course):
            return {"subject": o.subject, "id": o.id, "name": o.name}
        return json.JSONEncoder.default(self, o)


class Course:
    def __init__(self, label) -> None:
        tokens = label.split('-')
        self.subject = tokens[0].split(' ')[0].strip()
        self.id = tokens[0].split(' ')[1].strip()
        self.name = tokens[1].strip()

    def __str__(self) -> str:
        return self.toJSON()

    def toJSON(self):
        return json.dumps(self, indent=2, cls=Encoder)

    def __key(self):
        return (self.subject, self.id, self.name)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, Course):
            return self.__key() == other.__key()
        return NotImplemented

    def __lt__(self, other):
        return self.id < other.id


if __name__ == '__main__':
    c1 = Course('CS 475 - Intro to Parallel Programming')
    c2 = Course('CS 475 - Intro to Parallel Programming')
    c5 = Course('CS 561 - Data Structures')
    c3 = Course('CS 261 - Data Structures')
    c4 = Course('CS 224 - Data Science')
    print(c1)
    print(c2)
    print(c3)

    print(hash(c1))
    print(hash(c2))

    print(c1 == c2)
    print(c1 == c3)
    print(c3 == c4)
    print(c3 < c4)

    s0 = {c1, c1}
    s1 = {c1, c1}
    print(s0 == s1)

    tx = tuple(s0)
    ty = tuple(s1)
    print(tx == ty)

    s2 = {c1, c2}
    s3 = {c2, c1}
    print(s2 == s3)

    t1 = tuple({c1, c2})
    t2 = tuple({c2, c1})
    print(t1 == t2)

    tx = tuple(s1)
    ty = tuple(s2)
    print(tx == ty)

    s4 = {c1, c3}
    s5 = {c3, c1}
    print(s4 == s5)

    tx = tuple(s4)
    ty = tuple(s5)
    print(tx == ty)

    t3 = tuple({c1, c3})
    t4 = tuple({c3, c1})
    print(t3 == t4)

    s6 = {c1, c5}
    s7 = {c1, c4}
    print(s6 == s7)

    t5 = tuple({c1, c5})
    t6 = tuple({c1, c4})
    print(t5 == t6)

    tx = tuple(s6)
    ty = tuple(s7)
    print(tx == ty)

    s = set()
    s.add(c4)
    s.add(c1)
    s.add(c3)
    s.add(c5)
    sorted_list = sorted(list(s), key=lambda x: (x.id, x.subject))
    print(json.dumps(sorted_list, indent=2, cls=Encoder))
    print(json.dumps(c1, indent=2, cls=Encoder))
    print(c1)

    a = []
    a.append(c4)
    a.append(c1)
    a.append(c3)
    a.append(c5)
    print(json.dumps(c1, indent=2, cls=Encoder))
    print(json.dumps(a, indent=2, cls=Encoder))
    print(json.dumps(c1, indent=2, cls=Encoder))
